<copyright> TESLA mu.Compound extensions.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: Compound.t,v 1.17 1998/07/27 15:25:45 tiggr Exp $</id>
    </copyright>

implementation class
Compound extension tesla: tesla
{
  <doc> Various flags.  </doc>
  const C_TYPED_EXPRS = 1;
  const LOOP_CONDITION_AFTER = 2;

  <doc> For use within {writeFields}, the indentation of each top
      expression.  </doc>
  static MutableByteString wf_indent;
}

end;

implementation instance
Compound extension tesla
{
  <doc> Various flags.  </doc>
  int c_flags;
}

void
  coarseUsage int f
       parent Compound c
	 body MethodBody m
{
  [super coarseUsage f parent c body m];
  int i, n = [expressions length];
  if (n > 0)
    {
      for (i = 0; i < n - 1; i++)
	[expressions[i] coarseUsage 0 parent self body m];
      [expressions[n - 1] coarseUsage f parent self body m];
    }
}

<doc> Pass to the {container} or issue an error.  </doc>
Loop
  enclosingLoop
{
  if (!container)
    [[[ParseTom error] print "no enclosing loop"] nl];
  else
    //
    = [Compound (container) enclosingLoop];
}

protected Cexpr (last)
  precompileElements Cblock parent
	      method Cblock body
{
  int i, n;

  if (variables_ordered != nil)
    for ({i = 0; n = [variables_ordered length];}; i < n; i++)
      [parent addDecl [[Ctomvar alloc] init variables_ordered[i]]];

  n = [expressions length];
  for (i = 0; i < n; i++)
    last = [expressions[i] precompile parent method body];
}

Cexpr (result)
  precompile Cblock parent
      method Cblock body
{
  Cblock b = [[Cblock alloc] initWithParent parent];

  if (expr_flags & EXPR_RHS != 0)
    result = [type temporaryCexpr parent];

  Cexpr last = [self precompileElements b method body];

  if (result != nil)
    [last compile b assignmentTo result];

  [parent add b];
}

<doc> If this compound has not been typed yet, first type the elements,
    getting the type from the {last_expression}.  Then pass to {super}.
    </doc>
(Expression, Cons) (resolved, possible_types)
  resolve (Cons, Type) (expected, convertible)
  context Meta context_type
  nested: (MutableIntArray, int) (indices, index)
{
  if (!type)
    [self resolveLastExpression (expected, convertible)
	  context context_type nested: (indices, index)];

  if (c_flags & C_TYPED_EXPRS == 0)
    {
      c_flags |= C_TYPED_EXPRS;
      int i, n = [expressions length];
      for (i = 0; i < n; i++)
	{
	  Expression e = expressions[i];
	  Expression f = [e resolve (nil, nil) context context_type];
	  if (e != f)
	    {
	      // No need to check for last_expression changing, since it
	      // was already fully typed.
	      // Thu May 28 13:13:32 1998, tiggr@gerbil.org
	      expressions[i] = f;
	    }
	}
    }

  = [super resolve (expected, convertible) context context_type
	   nested: (indices, index)];
}

<doc> Resolve the last expression in the receiving compound.  </doc>
Expression
  resolveLastExpression (Cons, Type) (expected, convertible)
		context Meta context_type
		nested: (MutableIntArray, int) (indices, index)
{
  int n = [expressions length];
  if (n > 0)
    {
      Expression e = expressions[n - 1];
      Expression f = [e resolveLastExpression (expected, convertible)
			context context_type nested: (indices, index)];
      if (e != f)
	expressions[n - 1] = last_expression = f;
    }
  if (!last_expression)
    {
      if (flag_warn_empty_compound)
	[[[ParseTom warning self] print "compound without value"] nl];
      last_expression = void_expr;
    }
  type = [last_expression force_type];
  = self;
}

<doc> Emit the expressions.  </doc>
OutputStream
  writeFields OutputStream s
{
  if (!wf_indent)
    wf_indent = [MutableByteString new];

  [wf_indent add ' '];
  [wf_indent add ' '];

  int i, n = [expressions length];
  for (i = 0; i < n; i++)
    s = [[s nl] print (wf_indent, Expression (expressions[i]))];
  [wf_indent truncate [wf_indent length] - 2];
  = s;
}

end;

/******************** MethodBody ********************/

implementation class
MethodBody extension tesla: tesla
{
  <doc> Various flags.  </doc>
  const BODY_USES_RETURN = 1;
}

end;

implementation instance
MethodBody extension tesla
{
  int mb_flags;
}

Cblock (b)
  precompile
{
  b = [[Cbody alloc] initWithBody self];
  [self precompileElements b method b];
}

<doc> Update the {BODY_USES_RETURN} bit in the {mb_flags}.  </doc>
void
  set_uses_return boolean b
{
  mb_flags = b ? mb_flags | BODY_USES_RETURN : mb_flags & ~BODY_USES_RETURN;
}

<doc> Return the {BODY_USES_RETURN} bit in the {mb_flags}.  </doc>
boolean
  uses_return
{
  = mb_flags & BODY_USES_RETURN != 0;
}

end;

/******************** Loop ********************/

implementation class
Loop: Compound
{
  <doc> Counter for labels to which continue can jump.  </doc>
  static int num_loop_labels;
}

end;

implementation instance
Loop
{
  <doc> Various control fields.  </doc>
  Expression initializer, condition, modifier;

  <doc> A temporary for the value of this loop.  </doc>
  public Cexpr cresult;

  <doc> The name of the label for continue, if set.  </doc>
  String label;
}

<doc> Designated initializer.  </doc>
id
  initWithControl (Expression, Expression, Expression) (i, c, m)
	container Context cxt
{
  (initializer, condition, modifier) = (i, c, m);
  = [super initWithContainer cxt];
}

void
  coarseUsage int f
       parent Compound c
	 body MethodBody m
{
  if (initializer != nil)
    [initializer coarseUsage f & ~(EXPR_RHS | EXPR_LHS) parent c body m];
  if (condition != nil)
    [condition coarseUsage f & ~EXPR_LHS | EXPR_RHS parent c body m];
  if (modifier != nil)
    [modifier coarseUsage f & ~(EXPR_RHS | EXPR_LHS) parent c body m];
  [super coarseUsage f parent c body m];
}

<doc> Return {self}.  </doc>
Loop (self)
  enclosingLoop
{
}

<doc> Return a (newly created) {label}.  </doc>
String
  continue_label
{
  if (!label)
    label = [[MutableByteString new] print ("_label_", ++num_loop_labels)];
  = label;
}

(Expression, Cons) (resolved, possible_types)
  resolve (Cons, Type) (expected, convertible)
  context Meta context_type
  nested: (MutableIntArray, int) (indices, index)
{
  if (c_flags & C_TYPED_EXPRS == 0)
    {
      if (initializer != nil)
	initializer = [initializer resolve (nil, nil) context context_type];
      if (modifier != nil)
	modifier = [modifier resolve (nil, nil) context context_type];
      if (condition != nil)
	condition = [condition resolve
		     ([Cons with (basic_types_from_kind[BT_BOOLEAN], nil)], nil)
			       context context_type];
    }

  = [super resolve (expected, convertible) context context_type
	   nested: (indices, index)];
}

Cexpr (result)
  precompile Cblock parent
      method Cblock method_body
{
  Cblock bb = [[Cblock alloc] initWithParent parent];

  if (expr_flags & EXPR_RHS != 0)
    {
      result = cresult = [type temporaryCexpr parent];
      Cexpr iv = [[type defaultValue nil] precompile parent method method_body];
      [parent add [[Casgn alloc] init (result, iv)]];
    }

  if (initializer != nil)
    [initializer precompile parent method method_body];

  Cblock cc;
  if (condition != nil)
    {
      cc = [[Cblock alloc] initWithParent bb];
      Cexpr ccc = [condition precompile cc method method_body];
      [cc add [[Cite alloc condition]
		init ([[Cmonop alloc] init ("!", ccc)],
		      [[Cstmt alloc] init "break"], nil)]];
    }

  if (cc != nil && ![self condition_after])
    [bb add cc];

  Cexpr body_result = [self precompileElements bb method method_body];
  if (result != nil)
    [bb add [[Casgn alloc body_result] init (result, body_result)]];

  if (label != nil)
    [bb add [[Ccode alloc self]
	      init [[MutableByteString new] print (label, ":")]]];

  if (modifier != nil)
    [modifier precompile bb method method_body];

  if (cc != nil && [self condition_after])
    [bb add cc];

  [parent add [[Cfor alloc] init bb]];
  cresult = nil;
}

void
  set_condition Expression c
pre
  !condition
{
  condition = c;
}

/******************** flags ********************/

boolean
  condition_after
{
  = c_flags & 1 << LOOP_CONDITION_AFTER != 0;
}

void
  set_condition_after boolean b
{
  if (b)
    c_flags |= 1 << LOOP_CONDITION_AFTER;
  else
    c_flags &= ~(1 << LOOP_CONDITION_AFTER);
}

end;
