<copyright> TESLA mu.Method extensions.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: Method.t,v 1.32 1998/07/24 15:01:50 tiggr Exp $</id>
    </copyright>

/******************** Method ********************/

implementation class
Method extension tesla: tesla

end;

implementation instance
Method extension tesla

<doc> Return {TRUE} if the receiving method is equal to the {other}
    method, up to the indicated level of equality.  </doc>
deferred boolean (result)
	   equal Method other
	  names: boolean name_parts_p = FALSE
	  types: boolean types_p = FALSE
  selectorTypes: boolean selector_types_p = FALSE
	 context Meta context_type
post
  self == other -> result;

<doc> Compare the receiving method with the {other} custom method, up to
    the desired level of equality.  The default implementation returns
    {TRUE} if all options are {FALSE}.  </doc>
protected boolean
  equalCustomMethod id other
	      names boolean name_parts_p
	      types boolean types_p
      selectorTypes boolean selector_types_p
	    context Meta context_type
{
  = !name_parts_p && !types_p && !selector_types_p;
}

<doc> Compare the receiving method with the {other} builtin method, up to
    the desired level of equality.  The default implementation returns
    {TRUE} if all options are {FALSE}.  </doc>
protected boolean
  equalBuiltinMethod id other
	       names boolean name_parts_p
	       types boolean types_p
       selectorTypes boolean selector_types_p
	     context Meta context_type
{
  = !name_parts_p && !types_p && !selector_types_p;
}

<doc> Return the type for the argument at the {index} in the invocation
    {parts}, given the {context_type}.  </doc>
deferred Type (result)
  formalArgumentType int index
		  in Indexed parts
	     context Meta context_type
post
  result != nil;

<doc> Return {TRUE} if the receiving method involves somehow the dynamic
    type.  The default implementation returns {FALSE}.  </doc>
boolean
  involvesDynamicType
{
}

end;

/******************** BuiltinMethod ********************/

implementation class
BuiltinMethod extension tesla: tesla

end;

implementation instance
BuiltinMethod extension tesla

<doc> Return {TRUE} if the single name part with arguments matches the
    name of the operator of this method.  </doc>
boolean
  namePartsMatch Indexed parts
     withActuals Indexed actuals
{
  return ([parts length] == 1 && actuals != nil && [actuals length] == 1
	  && operator_names[operator] == String (parts[0]));
}

<doc> Pass to the {other} {equalBuiltinMethod names types selectorTypes}.
    </doc>
boolean
	   equal Method other
	  names: boolean name_parts_p = FALSE
	  types: boolean types_p = FALSE
  selectorTypes: boolean selector_types_p = FALSE
	 context Meta context_type
{
  = [other equalBuiltinMethod self names name_parts_p types types_p
	   selectorTypes selector_types_p context context_type];
}

<doc> Compare the receiving builtin method with the {other} buitlin
    method, up to the desired level of equality.  </doc>
boolean
  equalBuiltinMethod id other
	       names boolean name_parts_p
	       types boolean types_p
       selectorTypes boolean selector_types_p
	     context Meta context_type
{
  if (name_parts_p && [other operator] != [self operator])
    return FALSE;

  if (types_p)
    {
      if (![[arg1 actualSelf context_type]
	     equal [[other arg1] actualSelf context_type]])
	return FALSE;

      Type a2 = [self arg2], oa2 = [other arg2];
      if (!a2 != !oa2 || (a2 != nil && ![[a2 actualSelf context_type]
					  equal [oa2 actualSelf context_type]]))
	return FALSE;
    }

  if (selector_types_p && !types_p)
    if ([return_type typeDescription] != [[other return_type] typeDescription]
	|| [arg1 typeDescription] != [[other arg1] typeDescription])
      return FALSE;
    else
      {
	Type a12 = [self arg2], a22 = [other arg2];

	if (!a12 != !a22
	    || (a12 != nil && [a12 typeDescription] != [a22 typeDescription]))
	  return FALSE;
      }

  = TRUE;
}

Type (result)
  formalArgumentType int index
		  in Indexed parts
	     context Meta context_type
pre
  index == 0
{
  = arg1;
}

end;

/******************** BuiltinBinop ********************/

implementation class
BuiltinBinop extension tesla: tesla

end;

implementation instance
BuiltinBinop extension tesla

Type (result)
  formalArgumentType int index
		  in Indexed parts
	     context Meta context_type
pre
  index < 2
{
  if (index == 1)
    return arg2;
  = [super formalArgumentType index in parts context context_type];
}

<doc> Return {TRUE} if the two name parts with arguments match the name of
    the operator of this method and a colon.  </doc>
boolean
  namePartsMatch Indexed parts
     withActuals Indexed actuals
{
  return ([parts length] == 2 && actuals != nil && [actuals length] == 2
	  && operator_names[operator] == String (parts[0])
	  && parts[1] == [UniqueString with ":"]);
}

end;

/******************** CustomMethod ********************/

implementation class
mu.CustomMethod extension tesla: tesla
{
  <doc> The one and only {main} method.  </doc>
  public static CustomMethod main;

  <doc> Boolean flags to store in the {int} {flags}, so we do not spend a
      whole {boolean} (the size of a byte) on each one.  </doc>
  const METHOD_DYNAMIC_IN = 1;
  const METHOD_DYNAMIC_OUT = 2;
}

<doc> Describe the method with the indicated vitals to the stream {s}.
    </doc>
OutputStream
     describe OutputStream s
       method (Indexed, Indexed) (name_parts, arguments)
  returnType: Type return_type = nil
{
  int an = !arguments ? 0 : [arguments length];
  int i, nn = [name_parts length];
  String leading = "";

  if (return_type != nil)
    {
      s = [s print return_type];
      leading = " ";
    }

  for (i = 0; i < nn; i++)
    {
      String np = name_parts[i];

      s = [s print (leading, np)];
      leading = " ";
      if (i < an)
	s = [s print (leading, [arguments[i] type])];
    }

  = s;
}

end;

implementation instance
mu.CustomMethod extension tesla
{
  <doc> The formal selector of this method.  </doc>
  Selector formal_selector;

  <doc> The body of the method.  </doc>
  MethodBody body;

  <doc> The flags.  </doc>
  int flags;
}

<doc> Compute and return the actual selector for an invocation of the
    receiving method, when passed the {actual_arguments} (expressions) and
    returning the {actual_return}.  If {actual_arguments} is {nil}, the
    selector is computed for the formal arguments (and the
    {METHOD_DYNAMIC_IN} and {METHOD_DYNAMIC_OUT} of the {flags} are set if
    needed).  </doc>
Selector (result)
  actualSelector (Array, Type) (actual_arguments, actual_return)
{
  MutableByteString td_name_in = [MutableByteString new];
  MutableByteString sel_name = [MutableByteString new];
  Type ar = !actual_return ? return_type : actual_return;
  TypeDescription td_in, td_out;
  boolean dyn_in, dyn_out;

  dyn_out = [ar collectTypesDescription sel_name];
  if (!actual_return && dyn_out)
    flags |= METHOD_DYNAMIC_OUT;
  td_out = [TypeDescription withName sel_name];
  // Naughty!
  if ([sel_name length] > 1)
    {
      [sel_name truncate 0];
      [sel_name print ('(', [td_out name], ')')];
    }

  int i, na = !arguments ? 0 : [arguments length], nm = [name_parts length];
  Indexed aa = !actual_arguments ? arguments : actual_arguments;
  MutableByteString ad = [MutableByteString new];

  for (i = 0; i < nm; i++)
    {
      [sel_name print ('_', ({String s = name_parts[i];}))];
      if (i < na)
	{
	  [ad truncate 0];

	  /* Use the formal type, except if it is the dynamic type, in
             which case we use the actual type (if supplied).  */
	  Type aat = [arguments[i] type];
	  if (aat == dynamic_type)
	    aat = [aa[i] type];
	  dyn_in |= [aat collectTypesDescription ad];

	  [td_name_in print ad];
	  // Naughty!
	  if ([ad length] == 1)
	    [sel_name print ('_', ad)];
	  else
	    [sel_name print ("_(", ad, ')')];
	}
    }
  if (dyn_in && !actual_arguments)
    flags |= METHOD_DYNAMIC_IN;

  td_in = [TypeDescription withName td_name_in];
  = [Selector withName sel_name types (td_in, td_out)];
}

<doc> Insert default arguments into the {actual_arguments}, where needed,
    for an invocation of the receiving method with the given {name_parts}.
    </doc>
void
  addDefaults MutableArray actual_arguments
  accordingTo Indexed actual_name_parts
	  for mu.Invocation inv
{
  int na = !arguments ? 0 : [arguments length], nfn = [name_parts length];
  int nan = [actual_name_parts length];
  String anp = actual_name_parts[0];
  int fi, ai;

  for (fi = 0; fi < nfn; fi++)
    {
      String fnp = name_parts[fi];
      if (fnp == anp)
	anp = ++ai == nan ? nil : actual_name_parts[ai];
      else
	[actual_arguments insert [arguments[fi] default_value] at fi];
    }

  if (ai != nan)
    [[[ParseTom error inv]
       print ("internal: ai (", ai, ") != nan (", nan, ")")] nl];
  int naa = !actual_arguments ? 0 : [actual_arguments length];
  if (na != naa)
    [[[ParseTom error inv]
       print ("internal: na (", na, ") != naa (", naa, ")")] nl];
}

<doc> Pass to the {other} {equalCustomMethod names types selectorTypes
    context}.  </doc>
boolean
	   equal Method other
	  names: boolean name_parts_p = FALSE
	  types: boolean types_p = FALSE
  selectorTypes: boolean selector_types_p = FALSE
	 context Meta context_type
{
  = [other equalCustomMethod self names name_parts_p types types_p
	   selectorTypes selector_types_p context context_type];
}

<doc> Compare the receiving custom method with the {other} custom method,
    up to the desired level of equality.  </doc>
boolean
  equalCustomMethod id other
	      names boolean name_parts_p
	      types boolean types_p
      selectorTypes boolean selector_types_p
	    context Meta context_type
{
  if (self == other)
    return TRUE;

  // tomCCC This fails for: ![name_parts equal [other name_parts]].
  // Tue Apr 28 22:24:50 1998, tiggr@gerbil.org
  if (name_parts_p && ![[other name_parts] equal name_parts])
    return FALSE;

  if (types_p)
    {
      if (![[return_type actualSelf context_type]
	     equal [[other return_type] actualSelf context_type]])
	return FALSE;

      Indexed other_a = [other arguments];
      if (!other_a != !arguments)
	return FALSE;
      if (arguments != nil)
	{
	  int i, n = [arguments length], no = [other_a length];
	  if (n != no)
	    return FALSE;

	  for (i = 0; i < n; i++)
	    {
	      Argument a = arguments[i];
	      Argument oa = other_a[i];

	      if (![[[a type] actualSelf context_type]
		     equal [[oa type] actualSelf context_type]])
		return FALSE;
	    }
	}
    }

  /* If the types match, the selector types match by definition.  */
  if (selector_types_p && !types_p)
    {
      Selector other_sel = [other formal_selector];
      Selector my_sel = [self formal_selector];

      if (name_parts_p)
	{
	  if (other_sel != my_sel)
	    return FALSE;
	}
      else if ([my_sel types_in] != [other_sel types_in]
	       || [my_sel types_out] != [other_sel types_out])
	return FALSE;
    }

  = TRUE;
}

<doc> Return the {formal_selector}, creating it if necessary.  </doc>
Selector
  formal_selector
{
  if (!formal_selector)
    formal_selector = [self actualSelector (nil, nil)];
  = formal_selector;
}

<doc> Return {TRUE} if this is not a deferred method or a redeclaration.
    </doc>
boolean
  hasImplementation
{
  = ![self isDeferred] && ![self isRedeclare];
}

// This could be precomputed...
// Fri May 15 17:54:28 1998, tiggr@gerbil.org
<doc> Return {TRUE} if the receiving method employs the dynamic type for
    an argument or as a return type.  </doc>
boolean
  involvesDynamicType
{
  if (return_type == dynamic_type)
    return TRUE;
  if (arguments != nil)
    {
      int i, n = [arguments length];
      for (i = 0; i < n; i++)
	{
	  Argument a = arguments[i];
	  if ([a type] == dynamic_type)
	    return TRUE;
	}
    }
}

<doc> Return {TRUE} if an method invocation with the name {parts} and
    {actuals} arguments can match the receiving method.  </doc>
boolean
  namePartsMatch Indexed parts
     withActuals Indexed actuals
{
  int mine, mn = [name_parts length];
  int other, on = [parts length];

  if (on > mn)
    return FALSE;

  int formal_n = !arguments ? 0 : [arguments length];
  int actual_n = !actuals ? 0 : [actuals length];
  boolean other_with_arg = other < actual_n;
  String op = parts[other];

  for (mine = 0; mine < mn; mine++)
    {
      String mp = name_parts[mine];
      boolean mine_with_arg = mine < formal_n;

      /* Compare the unique strings with `=='.  If this is the last one,
	 and the arguments to not match, it is a fail (`init' does not
	 match `init int').  Additionally, make sure that a method `init
	 int a = 1 init' matches an invocation as `init'.  */
      if (op == mp && mine_with_arg == other_with_arg)
	{
	  op = ++other == on ? nil : parts[other];
	  other_with_arg = other < actual_n;
	}
      else if (!mine_with_arg || [arguments[mine] default_value] == nil)
	return NO;
    }

  = !op;
}

<doc> Resolve references, etc.  </doc>
void
  singleUnitLoaded int pass
{
  Meta context = [self meta];

  if (pass == 0)
    {
      if (arguments != nil)
	{
	  int i, n = [arguments length];
	  for (i = 0; i < n; i++)
	    [arguments[i] resolveDefaultValue context];
	}

      if ([self formal_selector] == load_selector && ![[ext meta] isInstance])
	[ext set_load_method self];
    }
  else if (pass == 1)
    {
      if (postcondition != nil)
	postcondition = [postcondition resolve
		     ([Cons with (basic_types_from_kind[BT_BOOLEAN], nil)], nil)
				       context context];

      if (precondition != nil)
	precondition = [precondition resolve
		     ([Cons with (basic_types_from_kind[BT_BOOLEAN], nil)], nil)
				     context context];

      if (body != nil)
	{
	  [TemporaryVariable reset];
	  [body typeAndResolve];
	  [body coarseUsage 0 parent nil body body];
	}

      if (verbose_post_single)
	{
	  [[[ParseTom report self]
	     print ("resolved ", [self formal_selector])] nl];
	  if (body != nil)
	    [[[ParseTom report self] print body] nl];
	}
    }
}

Type (result)
  formalArgumentType int index
		  in Indexed parts
	     context Meta context_type
{
  int i, n = [name_parts length];
  for (i = index; i < n; i++)
    if (String (parts[index]) == name_parts[i])
      return [[arguments[i] type] actualSelf context_type];

  /* This should not happen (and we return NIL).  */
}

<doc> Compile this method.  </doc>
TeslaOutput (of)
  writeCFile TeslaOutput of
{
  [formal_selector noteUsage];

  if (![self hasImplementation])
    return;

  [[return_type tupleFirstElement]
    writeCReturnType [[of nl self] print !body ? "extern " : ""]];
  [self writeCFunctionName [of nl]];

  if (!body)
    [of print "();"];
  else
    {
      [[[self meta] writeCReferenceTypeName [of print " ("]]
	print " self, selector cmd"];

      if (arguments != nil)
	{
	  int i, n = [arguments length];
	  for (i = 0; i < n; i++)
	    {
	      Argument a = arguments[i];
	      [a writeCDeclaration of prefix ", "];
	    }
	}
      if (return_names != nil)
	[return_names writeCFile of prefix ", " infix " *_r"
		      postfix nil skipFirst YES];

      [of print ')'];
      [[body precompile] writeCFile of];
    }
}

TeslaOutput (s)
  writeCFunctionName OutputStream s
{
  [formal_selector writeCName [[ext writeCNameSuffix s] print '_']];
}

TeslaOutput (s)
  writeCMethodDescription TeslaOutput s
{
  if (![self hasImplementation])
    return;
  [self writeCFunctionName [s print "\n{(int_imp) "]];
  [[ext meta] writeCMetaDefinitionName [s print ", (void *) &"]];
  [formal_selector writeCReferenceName [s print ", &"]];
  [s print "},"];
}

/******************** redefinitions ********************/

<doc> Be informed of the end of the body for the receiving method.  The
    closing brace has just been read.  </doc>
redefine void
  finishMethodBody
{
}

<doc> Be informed of the end of a method declaration.  The final semicolon
    has just been read.  </doc>
redefine void
  finishMethodDecl
{
  [self finishMethodHeading];

  if (![self isDeferred] && ![self isExtern] && ![self isRedeclare])
    [[[ParseTom error self]
       print "non-deferred non-extern non-redeclare method without body"] nl];
}

<doc> Return a reference to the {return_names}.  </doc>
Expression
  return_value
{
  = [return_names reference nil];
}

<doc> Be informed of the start of the body for the receiving method.  The
    opening brace has just been read.  </doc>
redefine MethodBody (compound)
  startMethodBody boolean not_really
pre
  !body
{
  [self finishMethodHeading];

  if (not_really)
    return;

  if (formal_selector == main_selector && ![[ext meta] isInstance])
    if (!main)
      main = self;
    else
      {
	[[[ParseTom error self]
	   print "multiple definitions of method main"] nl];
	[[[ParseTom error main] print "main previously defined here"] nl];
      }

  if ([self isDeferred])
    [[[ParseTom error] print "definition of method declared deferred"] nl];
  else if ([self isExtern])
    [[[ParseTom error] print "definition of method declared extern"] nl];
  else if ([self isRedeclare])
    [[[ParseTom error] print "definition of method redeclaration"] nl];
  else
    {
      body = compound = [[MethodBody alloc] initWithContainer self];

      /* Create names for the return values if necessary.  */
      if (!return_names && return_type != void_type)
	{
	  return_names = [ArgumentTuple withVariableTuple
					[return_type temporaryTuple body]];
	  [return_names updateLocation self];
	}
    }
}

/******************** private methods ********************/

<doc> Handle the stuff that is common for both a method declaration and a
    method definition.  </doc>
protected void
  finishMethodHeading
pre
  !formal_selector
{
  // If this is a redeclaration, we can fetch this stuff from the
  // overridden method, instead of reconstructing it...
  // Thu Apr 23 17:06:15 1998, tiggr@gerbil.org
  [self formal_selector];
  [ext addMethod self];
}

end; 
