<copyright> TESLA tesla.Selector.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: Selector.t,v 1.9 1998/07/22 12:54:26 tiggr Exp $</id>
    </copyright>

/******************** Selector ********************/

implementation class
Selector: State, tesla
{
  <doc> Various selector flags.  </doc>
  const SEL_DYN_IN = 1;
  const SEL_DYN_OUT = 2;
  const SEL_USED = 4;

  <doc> All selectors, keyed on their name.  </doc>
  static MutableDictionary selectors;

  <doc> All _used_ selectors.  </doc>
  public static MutableEqSet used_selectors;

  <doc> All used formal dynamic selectors.  </doc>
  public static MutableEqSet dynamic_formals;

  <doc> The number of dynamic selector map entries.  </doc>
  public static int num_dynamic_map_entries;
}

<doc> Create the {selectors}, to avoid checks in the other class methods.
    </doc>
void
  load Array arguments
{
  selectors = [MutableDictionary new];
  used_selectors = [MutableEqSet new];
  dynamic_formals = [MutableEqSet new];
}

<doc> Return the selector with the {name}.  If a selector by that name is
    not present in the known {selectors}, {nil} is returned.  </doc>
instance (id)
  named String name
{
  = selectors[name];
}

<doc> Return the selector with the indicated name parts and types.  </doc>
instance (id)
   withNameParts Indexed name_parts
	   types (Type, Indexed) (return_type, arg_types)
{
  MutableByteString td_name_in = [MutableByteString new];
  MutableByteString sel_name = [MutableByteString new];
  TypeDescription td_in, td_out;

  [return_type collectTypesDescription sel_name];
  td_out = [TypeDescription withName sel_name];
  if ([sel_name length] > 1)
    {
      [sel_name truncate 0];
      [sel_name print ('(', [td_out name], ')')];
    }

  int i, na = !arg_types ? 0 : [arg_types length], nm = [name_parts length];
  MutableByteString ad = [MutableByteString new];

  for (i = 0; i < nm; i++)
    {
      [sel_name print ('_', ({String s = name_parts[i];}))];
      if (i < na)
	{
	  Type at = arg_types[i];

	  [ad truncate 0];
	  [at collectTypesDescription ad];

	  [td_name_in print ad];
	  if ([ad length] == 1)
	    [sel_name print ('_', ad)];
	  else
	    [sel_name print ("_(", ad, ')')];
	}
    }

  td_in = [TypeDescription withName td_name_in];
  = [self withName sel_name types (td_in, td_out)];
}

<doc> Return the selector with the name and indicated argument and return
    types.  The type descriptions are used when creating the selector if
    it does not already exist.  The {name} is uniqued when creating a
    selector.

    The precondition protects against clashing method names like {void
    do_i_it int} and {void do int it int}.  </doc>
instance (id) (result)
  withName String name
     types (TypeDescription, TypeDescription) (td_in, td_out)
// gi can't handle this...
// Fri Apr 24 12:23:03 1998, tiggr@gerbil.org
// pre
//   ({
//     Selector sel = [self named[name]];
//     sel != nil -> ([td_in equal [sel types_in]]
// 		   && [td_out equal [sel types_out]]);
//   })
{
  result = selectors[name];
  if (!result)
    {
      result = [[self alloc] initWithName [UniqueString with name]
			     types (td_in, td_out)];
      selectors[name] = result;
    }
}

end;

implementation instance
Selector
{
  <doc> The name of this selector.  </doc>
  public String name;

  <doc> The description of the argument types.  </doc>
  public TypeDescription types_in;

  <doc> The description of the return type.  </doc>
  public TypeDescription types_out;

  <doc> Any actual selectors for this dynamic selector.  </doc>
  MutableArray dynamic_actuals;

  <doc> Various flags.  </doc>
  int sel_flags;
}

<doc> Designated initializer.  </doc>
protected id
  initWithName String n
	 types (TypeDescription, TypeDescription) (td_in, td_out)
{
  (name, types_in, types_out) = (n, td_in, td_out);

  if ([types_in includesDynamic])
    sel_flags |= SEL_DYN_IN;
  if ([types_out includesDynamic])
    sel_flags |= SEL_DYN_OUT;

  = [super init];
}

<doc> Return {TRUE} if the receiving selector employs a {dynamic}
    argument.  </doc>
boolean
  hasDynamicArguments
{
  = sel_flags & SEL_DYN_IN != 0;
}

<doc> Return {TRUE} if the receiving selector employs a {dynamic}
    return type.  </doc>
boolean
  hasDynamicReturn
{
  = sel_flags & SEL_DYN_OUT != 0;
}

<doc> Add the receiving selector to the {used_selectors}.  </doc>
void
  noteUsage
{
  if (sel_flags & SEL_USED == 0)
    {
      sel_flags |= SEL_USED;
      [used_selectors add self];
      [types_in noteUsage];
      [types_out noteUsage];
    }
}

void
  noteUseAsFormalFor Selector actual_selector
pre
  0 != sel_flags & (SEL_DYN_IN | SEL_DYN_OUT)
{
  if (!dynamic_actuals)
    {
      dynamic_actuals = [MutableObjectArray with actual_selector];
      [dynamic_formals add self];
    }
  else if ([dynamic_actuals indexOfIdentical actual_selector] < 0)
    [dynamic_actuals add actual_selector];
  [self noteUsage];
}

<doc> Also emit the {name} (that should suffice).  </doc>
OutputStream
  writeFields OutputStream s
{
  = [[super writeFields s] print (" name=", name)];
}

<doc> Declare the receiving selector.  </doc>
OutputStream (s)
  writeCDeclaration OutputStream s
{
  [[[self writeCReferenceName [s print "extern struct trtd_selector "]]
     print ';'] nl];
}

<doc> Define the receiving type description.  </doc>
OutputStream (s)
  writeCDefinition OutputStream s
{
  [[self writeCReferenceName [s print "struct trtd_selector "]]
    print (" =\n {0, {\"", name, "\", ", [name length], "}, &")];
  [[[types_out writeCReferenceName [[types_in writeCReferenceName s]
				     print ", &"]] print "};"] nl];
}

<doc> Emit a few lines for the dynamic map.  </doc>
OutputStream (s)
  writeCDynamicMap OutputStream s
pre
  dynamic_actuals != nil
{
  int i, n = [dynamic_actuals length];
  for (i = 0; i < n; i++)
    {
      Selector actual = dynamic_actuals[i];
      [actual writeCReferenceName [s print "\n{&"]];
      [[self writeCReferenceName [s print ", &"]] print "},"];
      num_dynamic_map_entries++;
    }
}

<doc> Emit something like {r_self}.  </doc>
OutputStream (s)
  writeCName OutputStream s
{
  [C write s CName name];
}

<doc> Emit the name of the struct containing the definition of this
    selector, i.e. something like {_sd_r_self} or {_sd_foo_r_self}.
    </doc>
OutputStream (s)
  writeCReferenceName OutputStream s
{
  [s print "_sd_"];
  if (single_unit != nil)
    [[C write s CName [single_unit name]] print '_'];
  [C write s CName name];
}

end; 

/******************** SelectorReference ********************/

implementation class
SelectorReference: Expression

end;

implementation instance
SelectorReference
{
  <doc> The selector.  </doc>
  Selector sel;
}

<doc> Designated initializer.  </doc>
id
  initWith Selector s
{
  sel = s;
  = [super initWithType selector_type];
}

Cexpr
  precompile Cblock parent
      method Cblock body
{
  [sel noteUsage];
  MutableByteString sel_ref = [[MutableByteString new] print "&"];
  [sel writeCReferenceName sel_ref];
  = [[Cconst alloc self] init sel_ref];
}

end;
