<copyright> TESLA mu.Type extensions.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: Type.t,v 1.22 1998/07/24 15:01:51 tiggr Exp $</id>
    </copyright>

/******************** Type ********************/

implementation class
Type extension tesla: tesla

<doc> Intersect the type lists {t1} and {t2}.  If {expected} is not nil,
    return a subsequence of {expected}.  </doc>
Cons (list)
  intersectTypes (Cons, Cons) (t1, t2)
	expected Cons expected
pre
  t1 != nil && t2 != nil
{
  Cons last;

  if (!expected)
    {
      // Quadratic!
      // Mon May 11 11:59:51 1998, tiggr@gerbil.org
      Type ty1, ty2;
      Cons t2l;

      while (t1 != nil)
	{
	  (ty1, t1) = [t1 decons];
	  for (t2l = t2; t2l != nil;)
	    {
	      (ty2, t2l) = [t2l decons];
	      Type ty = [ty1 bestFormalType ty2];
	      if (ty != nil)
		if (!list)
		  list = last = [Cons with (ty, nil)];
		else if (![list member ty])
		  {
		    Cons c = [Cons with (ty, nil)];
		    [last set_cdr c];
		    last = c;
		  }
	    }
	}
    }
  else
    {
      Type te, tt1 = [t1 car], tt2 = [t2 car];

      /* Walk EXPECTED, removing types from T1 and T2 as we encounter them
         (which is possible since they are in the same order), adding to
         the result when they are equal.  */
      while (expected != nil && t1 != nil && t2 != nil)
	{
	  (te, expected) = [expected decons];
	  if ([te equal tt1])
	    {
	      (tt1, t1) = [t1 decons];
	      if ([te equal tt2])
		{
		  Cons c = [Cons with (te, nil)];
		  if (!list)
		    list = c;
		  else
		    [last set_cdr c];
		  last = c;
		}
	    }
	  else if ([te equal tt2])
	    (tt2, t2) = [t2 decons];
	}
    }
}

end;

implementation instance
Type extension tesla

<doc> Return the actual self in the context of the {context} class or
    instance.  The default implementation returns {self}.  </doc>
id
  actualSelf Meta context
{
  = self;
}

<doc> Return the {type_list}, after having added the receiving type when
    none of the types already on the list are {equal} to the receiving
    type.  </doc>
Cons
  addToTypes Cons type_list
{
  Cons l = type_list;
  Type t;

  while (l != nil)
    {
      (t, l) = [l decons];
      if ([self equal t])
	return type_list;
    }

  = [Cons with (self, type_list)];
}

<doc> Return the {type_list} after having added those types from the
    {expected_list} that can match the receiving type.  </doc>
Cons
  addToTypes Cons type_list
     ordered Cons expected_list
     nested: (MutableIntArray, int) (indices, index)
pre
  expected_list != nil
{
  Cons previous = nil, current = type_list;
  Type et;

  while (expected_list != nil)
    {
      (et, expected_list) = [expected_list decons];
      if (current != nil && et == [current car])
	(previous, current) = (current, [current cdr]);
      else if ([self distanceToFormal [et typeAt index in indices]] >= 0)
	{
	  /* Not in there.  Append it to PREVIOUS, prepending it to
	     CURRENT.  */
	  Cons c = [Cons with (et, current)];
	  if (!previous)
	    type_list = c;
	  else
	    [previous set_cdr c];
	  previous = c;
	}
    }

  if (current != nil)
    [self unimplemented cmd message: "!!current"];

  = type_list;
}

<doc> Similar to {bestFormalType}, but now it is known that the {other} is
    an object.  The default implementation returns {nil}.  </doc>
Type
  bestFormalMeta Meta other
{
}

// Common superclass?
// Mon May 11 11:48:41 1998, tiggr@gerbil.org
<doc> Given the receiving and the {other} type, return the best matching
    formal type assuming that both types involved are actual types.  For
    non-object unequal types, this returns {nil}; for two object types,
    this returns the one that is the superclass of the other.  </doc>
Type
  bestFormalType Type other
{
  /* This is matchesExactly: in tomc.  */
  if (self == other)
    = self;
}

<doc> Append a type description (one or more of {TD_*} constants) to the
    {description}.  Return {TRUE} iff the dynamic type is involved.
    </doc>
deferred boolean
  collectTypesDescription MutableByteString description;

<doc> Return an expression for the default value of the receiving type.
    The default implementation returns 0 in some disguise.  </doc>
Expression
  defaultValue Common localizer
{
  = [[ForeignExpression alloc localizer] initWithExpression "0" type self];
}

<doc> Like {distanceToFormal}, but with {self} (now {actual}) and the
    {formal} (now {self}) swapped, and with the {actual} known to be a
    basic type.  The default implementation returns -1.  </doc>
protected int (result)
  distanceFromBasicType BasicType actual
{
  = -1;
}

<doc> Like {distanceToFormal}, but with {self} (now {actual}) and the
    {formal} (now {self}) swapped, and with the {actual} known to be a
    meta.  The default implementation returns -1.  </doc>
protected int (result)
  distanceFromMeta Meta actual
{
  = -1;
}

<doc> Like {distanceToFormal}, but with {self} (now {actual}) and the
    {formal} (now {self}) swapped, and with the {actual} known to be a
    type tuple.  The default implementation returns -1.  </doc>
protected int (result)
  distanceFromTypeTuple TypeTuple actual
{
  = -1;
}

<doc> Return the distance from the receiving type to the {formal} type.
    This returns an indication of the amount of conversion that is needed
    to feed an actual argument of the receiving type to a formal argument
    of the {formal} type.  Return -1 for infinite distance (impossible
    conversion); return {>= 0} for finite distance (doable conversion).
    </doc>
deferred int (result)
  distanceToFormal Type formal
post
  ([formal equal self] || formal == dynamic_type
   || formal == any_type) -> !result;

<doc> Return that element from {expected} that fits the receiving type
    best.  If {covariant_p}, the receiving type is the type of an actual,
    and {expected} contains the formal types.  If {!covariant_p}, the
    receiving type is the formal, and {expected} contains the possible
    actual types.  </doc>
id
  elementOf Cons expected
  covariant boolean covariant_p
    context Meta context_type
    nested: (MutableIntArray, int) (indices, index) = (nil, -1)
// tomCCC bogus code emitted?
// Mon May 11 22:01:46 1998, tiggr@gerbil.org
// pre
//   (index < 0 == !indices
//    && self == [self actualSelf context_type])
{
  Type t, it, match;
  int level = -1;

  while (expected != nil)
    {
      (t, expected) = [expected decons];
      it = [t typeAt index in indices];

      /* Best possible match.  */
      if (self == it)
	return t;

      /* Impossible match.  */
      if (!it)
	continue;

      int i = (covariant_p ? [self distanceToFormal it]
	       : [it distanceToFormal self]);
      if (i >= 0 && (!match || i < level))
	(match, level) = (t, i);
    }

  = match;
}

<doc> Return {TRUE} if the receiving type is a super type of the {meta}.
    The default implementation returns {FALSE}.  </doc>
boolean
  isSuperTypeOf Meta meta
{
}

<doc> Return {TRUE} if the receiving type is a tuple type of {num}
    elements.  Default implementation returns {FALSE}. </doc>
boolean
  isTupleOf int num
pre
  num > 1
{
}

<doc> Return {TRUE} iff the receiving type can be passed as an actual to
    the {formal} type, i.e., if it matches widening/covariantly.  The
    default implementation only returns {TRUE} for identical types.
    Subclasses should override.  </doc>
boolean
  matchesFormalType id formal
{
  if (self == formal)
    return TRUE;
}

<doc> Return a temporary variable of the receiving type.  </doc>
deferred Variable
  temporary Compound context;

<doc> Return a temporary Cexpr of the receiving type.  </doc>
deferred Cexpr
  temporaryCexpr Cblock parent;

<doc> Return a (possibly singleton) tuple of temporaries for the receiving
    type.  </doc>
deferred VariableTuple
  temporaryTuple Compound context;

<doc> Return the type at the indicated index.  The default implementation
    only returns {self} if the {index == -1}.  </doc>
Type
  typeAt int index
   from: int start_index = 0
      in IntArray indices
pre
  !indices == (index == -1)
{
  if (!indices)
    = self;
}

<doc> Return the first element of the receiving tuple of....  The default
    implementation returns {self}.  </doc>
Type
  tupleFirstElement
{
  = self;
}

<doc> Return the type description of the receiving type.  </doc>
deferred TypeDescription
  typeDescription;

<doc> Emit the name of the receiving type to the C output stream {s}.  </doc>
deferred OutputStream
  writeCType OutputStream s;

<doc> Like {writeCType}, but do it for the return type of the C
    function implementing a method.  The default implementation simply
    invokes {writeCType}, since most types do not need to discern
    between these two cases.  </doc>
OutputStream
  writeCReturnType OutputStream s
{
  = [self writeCType s];
}

<doc> Write a list of C types for the receiving type.  When
    {skip_first}, the first C type should be skipped, i.e. the first
    element of a tuple (transitively) should not be emitted.  This is
    used for emitting the type of a function in a function call.
    </doc>
deferred TeslaOutput (s)
  writeCTypes TeslaOutput s
       prefix String prefix
      postfix String postfix
    skipFirst boolean skip_first;

end;

/******************** TypeSingle ********************/

implementation class
TypeSingle extension tesla: tesla

end;

implementation instance
TypeSingle extension tesla

<doc> Return a new temporary.  </doc>
TemporaryVariable (tv)
  temporary Compound context
{
  tv = [TemporaryVariable withType self in context];
  [context addVariable tv];
}

<doc> Return a {Cref} to a {Cvar}.  </doc>
Cexpr
  temporaryCexpr Cblock parent
{
  Cvar cv = [parent temporary self];
  = [cv reference nil];
}

<doc> Return a temporary in a tuple.  </doc>
VariableTuple
  temporaryTuple Compound context
{
  = [VariableTuple with [self temporary context]];
}

TeslaOutput (s)
  writeCTypes TeslaOutput s
       prefix String prefix
      postfix String postfix
    skipFirst boolean skip_first
{
  if (!skip_first)
    {
      if (prefix != nil)
	[s print prefix];
      [self writeCType s];
      if (postfix != nil)
	[s print postfix];
    }
}

end;

/******************** TypeTuple ********************/

implementation class
TypeTuple extension tesla: tesla, Math

end;

implementation instance
TypeTuple extension tesla

// Possibly expensive...  could cache...
// Fri May 15 12:29:38 1998, tiggr@gerbil.org
<doc> Return the actual self in the context of the {context} class or
    instance.  If any of the elements of this tuple return something
    different, a new tuple is created and returned.  </doc>
id
  actualSelf Meta context
{
  int i;
  for (i = 0; i < length; i++)
    {
      Type t = self[i];
      Type at = [t actualSelf context];
      if (t != at)
	{
	  id r = [self copy];
	  r[i] = at;
	  for (i++; i < length; i++)
	    {
	      t = self[i];
	      at = [t actualSelf context];
	      if (t != at)
		r[i] = at;
	    }
	  return r;
	}
    }
  = self;
}

<doc> Pass to the elements of this tuple, and return {TRUE} if any of them
    does.  </doc>
boolean (has_dynamic)
  collectTypesDescription MutableByteString description
{
  int i, n = [self length];
  for (i = 0; i < n; i++)
    has_dynamic |= [self[i] collectTypesDescription description];
}

<doc> Return a tuple with default values.   </doc>
ExpressionTuple (result)
  defaultValue Common localizer
{
  int i;
  result = [[ExpressionTuple alloc localizer] initWithCapacity length];
  for (i = 0; i < length; i++)
    [result add [self[i] defaultValue localizer]];
}

<doc> Return the square root of the sum of squares of the distances of the
    elements of the {actual} to the elements of {self}.  Return -1 if any
    of the elements returns -1.  </doc>
int (result)
  distanceFromTypeTuple TypeTuple actual
{
  if (self == actual)
    return 0;

  int a_length = [actual length];
  if (a_length != length)
    return -1;

  int i, sq;
  for (i = 0; i < length; i++)
    {
      Type f = self[i];
      Type a = actual[i];
      int d = [a distanceToFormal f];
      if (d < 0)
	return -1;
      sq += d * d;
    }

  // rounding...
  // Thu May  7 11:38:14 1998, tiggr@gerbil.org
  return int (sqrt (double (sq)));
}

<doc> Tell the {other} to compute the {distanceFromTuple} with respect
    to {self}.  </doc>
int (result)
  distanceToFormal Type formal
{
  = (formal == dynamic_type || formal == any_type ? 0
     : [formal distanceFromTypeTuple self]);
}

<doc> Return {TRUE} if {num} matches the {length} of the receiving tuple.
    </doc>
boolean
  isTupleOf int num
{
  = num == length;
}

<doc> Pass this to the elements of the tuple type, and return a collection
    of their responses.  </doc>
VariableTuple (vt)
  temporary Compound context
{
  int i;

  vt = [VariableTuple withCapacity length];
  for (i = 0; i < length; i++)
    [vt add [self[i] temporary context]];
}

<doc> Return a {Ctuple} of temporaries.  </doc>
Ctuple (result)
  temporaryCexpr Cblock parent
{
  int i;

  result = [[Ctuple alloc] init length];
  for (i = 0; i < length; i++)
    [result add [self[i] temporaryCexpr parent]];
}

<doc> Pass to the first element.  </doc>
Type
  tupleFirstElement
{
  = [self[0] tupleFirstElement];
}

<doc> Retrieve the type at the indicated location, or pass to it.  </doc>
redefine Type
  typeAt int index
   from: int start_index
      in IntArray indices
{
  if (index < 0)
    return self;

  int my_index = indices[start_index];
  if (my_index >= length)
    return nil;

  Type elt = self[my_index];
  if (start_index == index)
    return elt;

  = [elt typeAt index from: start_index + 1 in indices];
}

<doc> Return the type description of the receiving tuple.  </doc>
TypeDescription
  typeDescription
{
  MutableByteString b = [MutableByteString withCapacity length];
  [self collectTypesDescription b];
  = [TypeDescription withName b];
}

<doc> Invoke {temporary}.  </doc>
VariableTuple (vt)
  temporaryTuple Compound context
{
  = [self temporary context];
}

TeslaOutput (s)
  writeCTypes TeslaOutput s
       prefix String prefix
      postfix String postfix
    skipFirst boolean skip_first
{
  int i;
  for (i = 0; i < length; i++)
    {
      [self[i] writeCTypes s prefix prefix postfix postfix
	   skipFirst skip_first];
      skip_first = FALSE;
    }
}

end;

/******************** BasicType ********************/

implementation class
BasicType: posing mu.BasicType, tesla
{
  <doc> Mapping from the basic kind {BT_*} thing to a type description
      {TD_*} thing.  </doc>
  static ByteArray basic_description;

  <doc> Mapping from a basic type to the C type name.  </doc>
  static Array basic_type_c_names;
}

<doc> Initialize {basic_description} and {basic_type_descriptions} before
    passing to {super}.  </doc>
void
  initialize boolean bare_environment
{
  // Should explicitly use TD_*...
  // Thu Apr 23 17:16:51 1998, tiggr@gerbil.org
  basic_description = "vobcilfdpsxrrr";

  int i;
  MutableArray a = [MutableObjectArray withCapacity BT_BEYOND - BT_FIRST];
  for (i = BT_FIRST; i < BT_BEYOND; i++)
    a[i] = [TypeDescription withName [[MutableByteString new]
				       print byte (basic_description[i])]];
  basic_type_descriptions = a;

  [TypeDescription registerNames [ObjectArray with
      ("TRT_TE_VOID", "TRT_TE_BOOLEAN", "TRT_TE_BYTE", "TRT_TE_CHAR",
       "TRT_TE_INT", "TRT_TE_LONG", "TRT_TE_FLOAT", "TRT_TE_DOUBLE",
       "TRT_TE_POINTER", "TRT_TE_SELECTOR", "TRT_TE_DYNAMIC",
       "TRT_TE_REFERENCE")]
		   ofDescriptions basic_type_descriptions];

  basic_type_c_names = [ObjectArray with ("void", "tom_byte", "tom_byte",
					  "tom_char", "tom_int", "tom_long",
					  "tom_float", "tom_double", "void *",
					  "selector", "dynamic", "void *",
					  "void *", "void *")];

  [super initialize bare_environment];
}

end;

implementation instance
BasicType

<doc> Eliminate an {BT_ID}, {BT_META_ID}, or {BT_ETAM_ID}.  </doc>
Type
  actualSelf Meta context
{
  if (kind >= BT_ID)
    if (kind == BT_ID)
      = context;
    else if (kind == BT_META_ID)
      = [context isClass] ? meta_meta : [context itsClass];
    else
      = ![context isClass] ? nil : [context instance];
  else
    = self;
}

<doc> Append the appropriate {TD_*} from the {basic_description}.  Return
    {TRUE} iff this is the dynamic type.  </doc>
boolean (result)
  collectTypesDescription MutableByteString description
post
  result -> self == dynamic_type
{
  byte b = basic_description[kind];
  = kind == BT_DYNAMIC;
  [description add b];
}

<doc> Return a positive number for widening integer or floating point
    conversions.  </doc>
int (result)
  distanceFromBasicType BasicType actual
{
  if (self == actual)
    return 0;

  int actual_kind = [actual kind];
  if (kind >= BT_BYTE && kind <= BT_DOUBLE
      && actual_kind >= BT_BYTE && actual_kind <= BT_DOUBLE)
    {
      if (kind <= BT_LONG)
	{
	  if (actual_kind <= BT_LONG && actual_kind <= kind)
	    return kind - actual_kind;
	}
      else if (actual_kind >= BT_FLOAT && actual_kind <= kind)
	return kind - actual_kind;
    }

  = -1;
}

<doc> Tell the {other} to compute the {distanceFromBasicType} with respect
    to {self}.  </doc>
int (result)
  distanceToFormal Type formal
{
  = (formal == dynamic_type || formal == any_type ? 0
     : [formal distanceFromBasicType self]);
}

<doc> Return {TRUE}.  </doc>
boolean
  dump_simple_p
{
  = TRUE;
}

<doc> For the dynamic type, return {self}, otherwise, pass to {super}.
    </doc>
Type
  typeAt int index
   from: int start_index
      in IntArray indices
{
  = self == dynamic_type ? self : [super typeAt index from: start_index
					 in indices];
}

<doc> Return appropriate one of {basic_type_descriptions}.  </doc>
TypeDescription
  typeDescription
{
  = basic_type_descriptions[kind];
}

OutputStream
  write OutputStream s
{
  = [s print ("BT:", [self name])];
}

OutputStream (s)
  writeCType OutputStream s
{
  [s print String (basic_type_c_names[kind])];
}

<doc> Suppress emitting anything resembling {void}.  </doc>
TeslaOutput (s)
  writeCTypes TeslaOutput s
       prefix String prefix
      postfix String postfix
    skipFirst boolean skip_first
{
  if (kind != BT_VOID)
    = [super writeCTypes s prefix prefix postfix postfix skipFirst skip_first];
}

OutputStream
  writeCReturnType OutputStream s
{
  if (self == dynamic_type)
    [s print "GENERIC_RETURN_TYPE"];
  else
    [self writeCType s];
}

end;

/******************** AnyType ********************/

implementation class
AnyType extension tesla: tesla

end;

implementation instance
AnyType extension tesla

end;

/******************** ErrorType ********************/

implementation class
ErrorType extension tesla: tesla

end;

implementation instance
ErrorType extension tesla

<doc> Return -1.  </doc>
int
  distanceToFormal Type formal
{
  = formal == dynamic_type || formal == any_type ? 0 : -1;
}

end;
