<copyright> TESLA ParseTom extensions.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: tom.t,v 1.32 1998/07/22 12:54:32 tiggr Exp $</id>
    </copyright>

implementation class
ParseTom: posing mu.ParseTom, tesla

end;

implementation instance
ParseTom
{
  <doc> Iff {TRUE}, the semantics of the current file are needed.  </doc>
  public boolean need_semantics;
}

<doc> Designated initializer.  </doc>
id
   init TomFile f
  lexer LexTom l
{
  need_semantics = TRUE;
  = [super init f lexer l];
}

<doc> Handle the /c/ token {body} and return the token to be returned (the
    next token).  </doc>
int
  cToken String body
      at int line
{
  ForeignCode e = [[ForeignCode alloc] initWithExpression body];
  [e set_line line];

  if (current_compound != nil)
    [current_compound addExpression e];
  else if (current_ext == nil)
    [file addForeignCode e];
  else
    [[ParseTom error e] print "<c> at inappropriate location"];

  = [lex nextToken];
}

/******************** overridden methods ********************/

redefine void
  addSuper Meta meta
     shift boolean instance_only
pre
  current_ext != nil
{
  if (instance_only)
    [current_ext addMeta meta];
  else
    {
      Meta mclass = [meta itsClass];
      String ename = [current_ext name];
      Meta i = ([current_meta isClass] ? Class ([current_meta instance])
		: current_meta);
      Meta c = [current_meta isClass] ? current_meta : [current_meta itsClass];
      Extension cext = [c extensionNamed ename];
      Extension iext = [i extensionNamed ename];
      [cext addMeta mclass];
      [iext addMeta meta];
      [self self];
    }
}

<doc> For local variables, return a reference.  Otherwise, return a proxy;
    we will search the meta later on.  This difference in handling is
    needed to preserve semantics of scope with the ordering of local
    variables.  </doc>
redefine Expression
  attributeNamed String name
{
  Variable v;
  if (current_compound != nil)
    v = [current_compound localEntityNamed name];
  if (!v)
    = [[ValueReferenceProxy alloc]
	initWithName name inContext current_meta];
  else
    = [v reference nil];
}

redefine Expression (binder)
  bind Bind binder
  body Expression body
{
  /* Drop the current compound, since it only served as a placeholder for
     the ConditionArgument.  */
  [binder set_body body];
  current_compound = Compound ([current_compound container]);
}

redefine Expression
  binop int op_token
   with (Expression, Expression) (lhs, rhs)
{
  if (op_token == OP_ASSIGN)
    = [[Assignment alloc] initWithLhs lhs rhs rhs];
  else if (op_token == OP_AT)
    = [[ArrayReference alloc] initWithArray lhs index rhs];
  else
    = [[UnresolvedOperatorInvocation alloc] initWithReceiver top_expression
	arguments [MutableObjectArray with (lhs, rhs)] operator op_token];
}

redefine Break
  break Expression argument
{
  = [[Break alloc nil] init (argument, [current_compound enclosingLoop])];
}

redefine Catch (catcher)
  catch Catch catcher
   body Expression body
{
  [catcher set_body body];
}

redefine Continue
  continue Expression argument
{
  = [[Continue alloc nil] init (argument, [current_compound enclosingLoop])];
}

redefine Loop (loop)
     do Loop loop
  while Expression condition
   body Expression body
{
  [loop set_condition condition];
  [self topExpression body];
  current_compound = Compound ([current_compound container]);
}

<doc> Pop the current compound.  </doc>
Expression
  finishCompound
{
  = current_compound;
  current_compound = Compound ([current_compound container]);
}

<doc> Tell this to the {method}.  </doc>
redefine void
  finishMethodBody CustomMethod method
{
  [method finishMethodBody];
  current_compound = nil;
  [lex set_skips FALSE];
  skips = FALSE;
}

<doc> Pass to the {method}.  </doc>
redefine void
  finishMethodDecl CustomMethod method
{
  [method finishMethodDecl];
}

Loop (for_loop)
   for Loop for_loop
  body Expression body
{
  [self topExpression body];
  current_compound = Compound ([current_compound container]);
}

<doc> Set up for reading a {bind}.  This involves making the {condition}
    argument visible.  </doc>
redefine void
  initBind
{
  Compound bc = [[Compound alloc] initWithContainer current_compound];
  ConditionArgument ca = [[ConditionArgument alloc]
			   initWithName [UniqueString with "condition"]
			   type condition_type container bc];
  [bc addVariable ca];
  current_compound = bc;
}

<doc> Set the {last_expression} of the {current_compound} and return
    {nil}.  </doc>
Expression (e)
  lastLocalVar LocalVariable variable
{
  [current_compound set_last_expression [variable reference nil]];
}

<doc> Not only add the local variable to the current compound; also emit
    an assignment for its default value.  </doc>
LocalVariable (lv)
  localVariable (Type, String, Expression) (t, n, e)
{
  if (current_compound != nil)
    {
      lv = [current_compound localVariable (t, n) for nil];
      [self topExpression [[Assignment alloc]
			    initWithLhs [lv reference nil]
			    rhs !e ? [[lv type] defaultValue nil] : e]];
    }
}

// Does it inherit the default values?
// Mon Apr 27 10:41:21 1998, tiggr@gerbil.org
redefine Expression
   monop int op_token
    with Expression shs
  postp: boolean postfix
{
  if (op_token == OP_RET || op_token == OP_SET)
    {
      ReturnAssignment ra = [ReturnAssignment alloc];
      ra = [ra initWithLhs (shs == void_expr ? shs
			    : [current_method return_value]) rhs shs];
      if (op_token == OP_RET)
	[ra set_with_jump TRUE];

      return ra;
    }

  = [[UnresolvedOperatorInvocation alloc]
      initWithReceiver top_expression arguments [MutableObjectArray with shs]
      operator op_token postfixed: postfix];
}

<doc> Be informed of the postcondition {e}.  </doc>
void
  postcondition Expression e
       relative int relative
       ofMethod CustomMethod m
{
  [m set_postcondition e];
  current_compound = nil;
}

<doc> Be informed of the precondition {e}.  </doc>
void
  precondition Expression e
      relative int relative
      ofMethod CustomMethod m
{
  [m set_precondition e];
  current_compound = nil;  
}

<doc> Return a {SelectorReference}.  </doc>
Expression
  selectorWithNameParts MutableArray name_parts
		  types (Type, MutableArray) (return_type, arg_types)
{
  = [[SelectorReference alloc]
      initWith [Selector withNameParts name_parts
  			 types (return_type, arg_types)]];
}

Bind
  startBind MutableArray condition_tuples
{
  = [[Bind alloc current_compound] initWithHandlers condition_tuples];
}

redefine Expression
  startCatch Expression what
{
  = [[Catch alloc what] initWithTag what];
}

<doc> Start the extension {ext}.  Since we do not read interface (.j)
    files, it should never be an interface.  </doc>
redefine void
  startExtension Extension ext
       interface boolean is_interface
pre
  !is_interface
{
  if ([ext implemented])
    {
      [[[self error] print ("redefinition of ", [ext descriptive_name])] nl];
      [[[self error ext]
	 print ([ext descriptive_name], " previously defined here")] nl];
    }
  else
    {
      [ext set_implemented TRUE];
      [ext set_line [lex current_line]];
      [ext set_containing_file [lex filename]];
    }
}

<doc> Push a new compound.  </doc>
void
  startCompound
{
  current_compound = [[Compound alloc] initWithContainer current_compound];
}

redefine Loop (do_loop)
  startDo
{
  current_compound = do_loop
    = [[Loop alloc] initWithControl (nil, nil, nil) container current_compound];
  [do_loop set_condition_after TRUE];
}

<doc> Update the {need_semantics} according to what the {file} tells us.
    </doc>
redefine void
  startFile
{
  need_semantics = [file need_semantics];
}

Loop (for_loop)
  startFor (Expression, Expression, Expression) (i, c, m)
{
  current_compound = for_loop
    = [[Loop alloc] initWithControl (i, c, m) container current_compound];
}

<doc> Set the {current_compound} to the value returned by telling the
    {method} to {startMethodBody}.  </doc>
redefine void
  startMethodBody CustomMethod method
{
  if (!need_semantics)
    {
      [lex set_skips TRUE];
      skips = TRUE;
    }
  current_compound = [method startMethodBody skips];

  if (need_semantics)
    {
      /* Initialize the non-argument return values.  */
      ArgumentTuple at = [method return_names];
      if (at != nil)
	{
	  Argument bt = [at initializeNonArgumentReturnValues current_compound];
	  if (at != bt)
	    [self unimplemented cmd message: "at != bt"];
	}
    }
}

<doc> Set up a {current_compound}.  </doc>
void
  startPostcondition
{
  current_compound = [[Compound alloc] initWithContainer current_method];
}

<doc> Set up a {current_compound}.  </doc>
void
  startPrecondition
{
  current_compound = [[Compound alloc] initWithContainer current_method];
}

redefine Unwind
  startUnwind Expression what
{
  = [[Unwind alloc what] initWithProtection what];
}

redefine Loop (while_loop)
  startWhile Expression c
{
  current_compound = while_loop
    = [[Loop alloc] initWithControl (nil, c, nil) container current_compound];
}

<doc> Add the {top_expression} to the {current_compound}, if it is not
    {nil} (which it is for local-variable declarations).  </doc>
void
  topExpression Expression top_expression
{
  if (top_expression != nil)
    {
      if (verbose_parse_pre)
	[[[ParseTom report] print ("*****: ", top_expression)] nl];
      [current_compound addExpression top_expression];
    }
}

redefine Expression
  triop int op_token
   with (Expression, Expression, Expression) (lhs, mhs, rhs)
pre
  op_token == OP_ITE
{
  = [[ITE alloc] initWithCondition lhs branches (mhs, rhs)];
}

redefine Unwind (unwinder)
  unwind Unwind unwinder
    body Expression body
{
  [unwinder set_body body];
}

Loop (while_loop)
  while Loop while_loop
   body Expression body
{
  [self topExpression body];
  current_compound = Compound ([current_compound container]);
}

end; 
